from copy import deepcopy
from time import sleep
from uuid import uuid4, UUID

from bson import Binary, binary

from cocktail_apikit import DictMongoQueryBuilder, MongoDBManager
from cocktail_apikit.utils_kit import build_mongodb_uuid, convert_py_uuid_to_mongodb_uuid


def populate_with_pets(db_client):
    db_client.create({'_id': build_mongodb_uuid(uuid4()), 'name': 'cat red'})
    db_client.create({'_id': build_mongodb_uuid(uuid4()), 'name': 'dog blue'})
    db_client.create({'_id': build_mongodb_uuid(uuid4()), 'name': 'duck yellow'})


def test_dict_mongo_query_builder_or_conditional(demo_schema):
    query_dict = dict()
    query_dict['or'] = [
        {"name": 'person name'},
        {"documents.number": '552244554789'}
    ]

    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)
    mongo_query = builder.to_mongo_query()
    mongo_query_dict = mongo_query.to_dict()

    assert mongo_query_dict['or'] == [{'name': 'person name'}, {'documents.number': '552244554789'}]


def test_dict_mongo_query_builder(demo_schema):
    id_value = str(uuid4())
    query_dict = {'name__eq': 'demo', 'id__ne': id_value}

    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)

    mongo_query = builder.to_mongo_query()

    print(mongo_query.to_dict())
    mongo_query_dict = mongo_query.to_dict()

    assert mongo_query_dict['name'] == {'$eq': 'demo'}
    assert mongo_query_dict['_id'] == {'$ne': Binary(UUID(id_value).bytes, binary.STANDARD)}

    for key in ['sort', 'projection', 'page', 'limit', 'skip']:
        assert key in mongo_query_dict


def test__mongo_db_manager__caching_connection_ok(mongo_uri):
    config1 = {'MONGODB_URI': mongo_uri, 'DB_NAME': 'testDB', 'COLLECTION_NAME': 'demos'}
    config2 = {'MONGODB_URI': mongo_uri, 'DB_NAME': 'testDB', 'COLLECTION_NAME': 'demos2'}
    config3 = {'MONGODB_URI': mongo_uri, 'DB_NAME': 'testDB', 'COLLECTION_NAME': 'demos3'}
    db1 = MongoDBManager(config1)
    db2 = MongoDBManager(config2)
    db3 = MongoDBManager(config3)

    assert db1.CLIENTS == db2.CLIENTS == db3.CLIENTS
    assert len(db1.CLIENTS) == len(db2.CLIENTS) == len(db3.CLIENTS) == 1
    assert db1.collection.name == 'demos'
    assert db2.collection.name == 'demos2'
    assert db3.collection.name == 'demos3'


def test_mongo_query_object_to_dict_no_in_place_update(demo_schema):
    id_value = str(uuid4())
    query_dict = {'name__eq': 'demo', 'id__ne': id_value}

    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)

    mongo_query = builder.to_mongo_query()
    query_condition = deepcopy(mongo_query.condition)
    mongo_query.to_dict()
    assert query_condition == mongo_query.condition


def test_filter_with_aggregate_find(db_client, demo_schema):
    populate_with_pets(db_client)
    query_dict = {'name': 'dog blue'}
    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)
    mongo_query = builder.to_mongo_query()
    results, count = db_client.filter_with_aggregate(mongo_query)
    pet_names = map(lambda pet: pet['name'], results)
    assert 'dog blue' in pet_names
    assert 'cat red' not in pet_names


def test_filter_with_aggregate_sort(db_client, demo_schema):
    populate_with_pets(db_client)
    query_dict = {'sort': ['-name']}
    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)
    mongo_query = builder.to_mongo_query()
    results, count = db_client.filter_with_aggregate(mongo_query)
    pet_names = list(map(lambda pet: pet['name'], list(results)))
    assert ['duck yellow', 'dog blue', 'cat red'] == pet_names


def test_filter_with_aggregate_limit(db_client, demo_schema):
    populate_with_pets(db_client)
    query_dict = {'limit': 1}
    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)
    mongo_query = builder.to_mongo_query()
    results, count = db_client.filter_with_aggregate(mongo_query)
    assert len(list(results)) == 1


def test_filter_with_aggregate_pagination(db_client, demo_schema):
    populate_with_pets(db_client)
    query_dict = {'page': 3, 'limit': 1}
    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)
    mongo_query = builder.to_mongo_query()
    results, count = db_client.filter_with_aggregate(mongo_query)
    assert list(results)[0]['name'] == 'duck yellow'


def test_filter_with_aggregate_text_search(db_client, demo_schema):
    populate_with_pets(db_client)
    db_client.collection.create_index([('name', 'text')])
    query_dict = {}
    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)
    mongo_query = builder.to_mongo_query()
    results, count = db_client.filter_with_aggregate(mongo_query, text_search='cat')
    result_list = list(results)
    assert result_list[0]['name'] == 'cat red'
    assert len(result_list) == 1


def test_filter_with_aggregate_relations(db_client, demo_schema):
    populate_with_pets(db_client)
    cat = list(db_client.find())[0]

    db_client.db.get_collection('pet_sounds').insert_one({'_id': convert_py_uuid_to_mongodb_uuid(cat['_id']),
                                                          'sound': 'Meow', '_enabled': True})
    query_dict = {
        '_id': cat['_id'],
        'lookup': {
            'from': 'pet_sounds',
            'localField': '_id',
            'foreignField': '_id',
            'as': 'sounds',
        },
        'unwind': '$sounds'
    }

    builder = DictMongoQueryBuilder(query_data=query_dict, schema=demo_schema)
    mongo_query = builder.to_mongo_query()
    results, count = db_client.filter_with_aggregate(mongo_query)
    list_results = list(results)
    assert list_results[0]['sounds']['sound'] == 'Meow'


def test_build_range_filter_with_schema_fields(demo_schema):
    query = {'name__in': 'hello,world',
             'id__in': '70dafa5e-d574-4b95-840e-1d14be8da7ad,212a1e2d-f674-44f5-9d8d-b2873647bf10'}
    builder = DictMongoQueryBuilder(query_data=query, schema=demo_schema)
    mongo_query = builder.to_mongo_query().to_dict()
    assert 'name' in mongo_query
    assert isinstance(mongo_query['name']['$in'], list)
    assert all([isinstance(o, str) for o in mongo_query['name']['$in']])

    assert '_id' in mongo_query
    assert isinstance(mongo_query['_id']['$in'], list)
    assert all([isinstance(o, Binary) and o.subtype == 4 for o in mongo_query['_id']['$in']])



import os
import subprocess

from setuptools import setup, find_packages


def read_file(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()


def git_version():
    """
    Fetch version from git information
    """
    git_default_version = '0.1.0'
    try:
        git_tag = subprocess.check_output(['git', 'describe', '--tags'])
        if git_tag:
            return git_tag.strip()[1:].decode('utf-8')
    except Exception:
        return git_default_version


setup(
    name="cocktail-apikit",

    version=git_version(),

    description="A collection of tools for APIs",
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',

    url='http://gitlab.com/theo-l/cocktail_apikit',

    author='Liang Guisheng',
    author_email='theol.liang@truckpad.com.br',

    license='MIT',

    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],

    keywords='cocktail_apikit backend apikit',

    packages=["cocktail_apikit"],

    install_requires=read_file('requirements.txt').strip().split('\n')[1:],

    extras_require={
        'dev': read_file('requirements-dev.txt').strip().split('\n')[2:]
    }

)



"""
Marshmallow tool kit will be used by backend api projects
"""
import uuid
from datetime import datetime, timezone
from decimal import Decimal

from bson import Binary
from marshmallow import Schema, pre_dump, post_load, post_dump, pre_load, \
    fields
from marshmallow.fields import DateTime, UUID

from .constants import (MARSHMALLOW_NESTED_FIELDS, MARSHMALLOW_LIST_FIELDS,
                        MARSHMALLOW_DICT_FIELDS,
                        MARSHMALLOW_SCALAR_FIELDS, MONGO_ID_FIELD_NAME,
                        POTENTIAL_MONGO_ID_FIELD_NAME)
from .utils_kit import convert_py_uuid_to_mongodb_uuid


############################################################
# Customized marshmallow fields
############################################################

class MongoUUID(fields.UUID):
    """
    Convert value between uuid string to Mongo UUID(bson.binary.Binary)
    """

    def _serialize(self, value, attr, obj):
        """
        Convert MONGO UUID (Binary) to string
        """
        value = uuid.UUID(bytes=value) if value and isinstance(value, Binary) else value
        return super()._serialize(value, attr, obj)

    def _deserialize(self, value, attr, data):
        """
        Convert str to Mongo UUID
        """
        value = super()._deserialize(value, attr, data)
        if not value:
            return value
        return convert_py_uuid_to_mongodb_uuid(value)


class IntegerTimeStamp(DateTime):
    """
    Convert value between int and datetime string
    """

    def _serialize(self, value, attr, obj):
        """
        Convert a **integer** timestamp value & **string** datetime  value
        """
        if isinstance(value, int):
            value = datetime.utcfromtimestamp(value)
        return super()._serialize(value, attr, obj)

    def _deserialize(self, value, attr, data):
        """
        Convert a **string** datetime  value to **integer** timestamp  value
        """
        value = super()._deserialize(value, attr, data)
        if not value:
            return None
        return int(value.timestamp())


class DecimalTimeStamp(DateTime):
    """
    Convert value between **Decimal** timestamp value and **string** datetime  value
    """

    def _serialize(self, value, attr, obj):
        """
        Convert a **Decial** timestamp value to **string** datetime  value
        """
        if isinstance(value, Decimal):
            value = datetime.utcfromtimestamp(float(value))
        return super()._serialize(value, attr, obj)

    def _deserialize(self, value, attr, data):
        """
        Convert a **string** datetime  value to **Decimal** timestamp value
        """
        value = super()._deserialize(value, attr, data)
        if not value:
            return None
        return Decimal(value.timestamp())


############################################################
# Customized Base Schema which will by used by all other api projects
############################################################
class SchemaMongoMixin:
    """
    This mixin class will help marshmallow schema class generate a collection of valid mongo-style query field names
    """

    @pre_dump
    def pre_dump_process(self, data):
        data['id'] = data.get(MONGO_ID_FIELD_NAME, None)
        data = self._extra_pre_dump_process(data)
        return data

    def _extra_pre_dump_process(self, data):
        """
        Any other extra pre dump process can be added here
        """
        return data

    @post_dump
    def post_dump_process(self, data):
        data = self._extra_post_dump_process(data)
        return data

    def _extra_post_dump_process(self, data):
        """
        Any other extra post dump process can be added here
        """
        return data

    @pre_load
    def pre_load_process(self, data):

        data = dict(data)

        # NOTE: 
        #   1. in create case, normally don't have 'id'/'_id' field
        #   2. in update case, if the 'id' field exists in payload, keep it instead of create new one 
        #   3. in rare case the '_id' field exists in payload, then pass it to 'id'

        if not data.get(POTENTIAL_MONGO_ID_FIELD_NAME, None) and not data.get(MONGO_ID_FIELD_NAME, None):
            data[POTENTIAL_MONGO_ID_FIELD_NAME] = uuid.uuid4().hex
        elif data.get(MONGO_ID_FIELD_NAME, None):
            data[POTENTIAL_MONGO_ID_FIELD_NAME] = data.pop(MONGO_ID_FIELD_NAME)

        data = self._extra_pre_load_process(data)
        return data

    def _extra_pre_load_process(self, data):
        """
        Any other extra pre load process can be added here
        """
        return data

    @post_load
    def post_load_process(self, data):

        # auto generate these field after load data from client input
        if 'created_at' not in data:
            data['created_at'] = datetime.now().astimezone(timezone.utc)

        if 'updated_at' not in data:
            data['updated_at'] = datetime.now().astimezone(timezone.utc)

        # always make sure mongo default id name '_id' in the loaded data 
        if data.get(POTENTIAL_MONGO_ID_FIELD_NAME, None):
            data[MONGO_ID_FIELD_NAME] = data.pop(POTENTIAL_MONGO_ID_FIELD_NAME)

        data = self._extra_post_load_process(data)
        return data

    def _extra_post_load_process(self, data):
        """
        Any other extra post load process can be added here
        """
        return data

    @classmethod
    def valid_mongo_query_fields(cls) -> dict:
        """
        Use class level cache Schema's valid mongo style's query  field names to avoid call the real method repeatedly
        :return: a set of valid query field name and its marshmallow type
        """
        if getattr(cls, '_valid_mongo_query_fields', False):
            return getattr(cls, '_valid_mongo_query_fields')

        def _fetch_valid_field_name_from_schema_fields(prefix: str = '', fields: dict = None):

            field_names = {}

            for field, value in fields.items():

                sub_prefix = '{}.{}'.format(prefix, field) if prefix else field

                # Nested field
                if isinstance(value, MARSHMALLOW_NESTED_FIELDS):
                    field_names.update(
                        _fetch_valid_field_name_from_schema_fields(sub_prefix, value.nested._declared_fields),
                    )
                    continue

                # List fields
                if isinstance(value, MARSHMALLOW_LIST_FIELDS):
                    if isinstance(value.container, MARSHMALLOW_NESTED_FIELDS):
                        try:
                            field_names.update(
                                _fetch_valid_field_name_from_schema_fields(sub_prefix, value.container.nested.fields))
                        except AttributeError:
                            continue
                        else:
                            continue

                # Dict fields
                if isinstance(value, MARSHMALLOW_DICT_FIELDS):
                    field_names.update(_fetch_valid_field_name_from_schema_fields(sub_prefix, value.metadata))
                    continue

                # Scalar value field
                if isinstance(value, MARSHMALLOW_SCALAR_FIELDS):
                    field_dump_to = value.dump_to
                    if not field_dump_to:
                        field_names.update({sub_prefix: type(value)})
                    else:
                        field_names.update(
                            {'.'.join(sub_prefix.split('.')[0:-1] + [field_dump_to]) + '->' + field: type(
                                value)})
                        field_names.update(
                            {'.'.join(sub_prefix.split('.')[0:-1] + [field_dump_to]): type(value)})

            return field_names

        valid_mongo_query_fields = _fetch_valid_field_name_from_schema_fields(fields=cls._declared_fields)

        # added mongo's default id field name '_id' to valid field set
        valid_mongo_query_fields.update({MONGO_ID_FIELD_NAME: UUID})
        setattr(cls, '_valid_mongo_query_fields', valid_mongo_query_fields)
        return getattr(cls, '_valid_mongo_query_fields')


class BaseSchema(Schema, SchemaMongoMixin):
    id = MongoUUID(required=False, allow_none=True, description='ID of a schema record')
    created_at = DateTime(required=False, description='Schema creation time')
    updated_at = DateTime(required=False, description='Schema object update time')
    deleted_at = DateTime(required=False, description='Schema object delete time', allow_none=True)



import random
import string
import uuid

from bson import Binary
from bson.binary import STANDARD


def test_schema_mongo_valid_fields(demo_schema):
    valid_mongo_query_fields = demo_schema.valid_mongo_query_fields()
    assert len(valid_mongo_query_fields) == 8


def test_decimal_timestamp_deserialize(demo_schema, valid_decimal_timestamp, valid_datetime_string):
    input_data = {
        'name': 'demo',
        'ts': valid_datetime_string
    }
    loaded_data, errors = demo_schema.load(input_data)
    assert errors == {}
    assert loaded_data['ts'] == valid_decimal_timestamp


def test_decimal_timestamp_serialize(demo_schema, valid_decimal_timestamp, valid_datetime_string):
    input_data = {
        'name': 'demo',
        'ts': valid_decimal_timestamp
    }
    dumped_data, errors = demo_schema.dump(input_data)
    assert errors == {}
    assert dumped_data['ts'] == valid_datetime_string


def test_mongo_uuid_field_deserialize_serialize(demo_schema):
    raw_data = {
        'name': 'test',
        'mongo_uuid': None
    }

    ############################################################
    # Test a valid None value
    ############################################################
    loaded_data, error = demo_schema.load(raw_data)

    assert not error
    assert loaded_data['mongo_uuid'] is None

    dumped_data, error = demo_schema.dump(loaded_data)
    assert not error
    assert dumped_data['mongo_uuid'] is None

    ############################################################
    # Test a valid uuid string
    ############################################################
    raw_data['mongo_uuid'] = uuid.uuid4().hex

    print(raw_data)
    loaded_data2, error = demo_schema.load(raw_data)

    assert not error
    assert isinstance(loaded_data2['mongo_uuid'], Binary) and loaded_data2['mongo_uuid'].subtype == STANDARD

    dumped_data2, error = demo_schema.dump(loaded_data2)
    assert not error
    assert isinstance(dumped_data2['mongo_uuid'], str) and dumped_data2['mongo_uuid'].replace('-', '') == raw_data[
        'mongo_uuid']

    ############################################################
    # Test an invalid value
    ############################################################
    raw_data['mongo_uuid'] = ''.join(random.sample(string.ascii_letters, random.randint(1, 31)))
    loaded_data3, error = demo_schema.load(raw_data)
    assert 'mongo_uuid' in error


def test_schema_load_without_id_will_auto_create_ok(demo_schema):
    """
    Mock creation case when create object no need to specify id field 
    """
    payload = {
        'name': 'test'
    }

    loaded_data, error = demo_schema.load(payload)
    assert not error
    assert '_id' in loaded_data


def test_schema_load_with_id_will_keep_the_same_value_ok(demo_schema):
    """
    Mock update case when update object which may contain nested object with id field
    this case keep the orignal id field value instead of create a new one
    """

    payload = {
        'name': 'test',
        'id': uuid.uuid4().hex  
    }
    loaded_data1, error = demo_schema.load(payload)
    assert not error
    assert '_id' in loaded_data1

    payload['_id'] = payload.pop('id')  # Test a rare case 
    loaded_data2, error = demo_schema.load(payload)
    assert not error
    assert '_id' in loaded_data2

    assert isinstance(loaded_data1['_id'], Binary) 
    assert isinstance(loaded_data2['_id'], Binary) 
    assert loaded_data1['_id'] == loaded_data2['_id']


def test_schema_load_with_empty_id_value_ok(demo_schema):
    payload = {
        'name': 'test',
        'id': ''
    }
    loaded_data, error = demo_schema.load(payload)

    assert not error

    assert '_id' in loaded_data and isinstance(loaded_data['_id'], Binary) and loaded_data['_id'].subtype == STANDARD


def test_schema_load_with_invalid_uuid_as_id_value_fail(demo_schema):
    payload = {
        'name': 'teset',
        'id': 'any invalid uuid value'
    }
    loaded_data, error = demo_schema.load(payload)
    assert 'id' in error and 'Not a valid UUID.' in error['id']
